import asyncio
import json
from typing import List, Type

from fastapi import WebSocket
from pydantic import BaseModel, root_validator
from zen_dash import ZenPage
from zen_dash.support.encoder import JsonEncoder


async def send_data(websocket: WebSocket, pages: List[ZenPage], myencoder=JsonEncoder):
    page_dict = {}
    for p in pages:
        if p.tab_number is None:
            raise Exception("Tab number is not define in Zen Page")
        
        if p.subtab_number is None:
            page_name = f"page_{p.tab_number}"
        else:
            page_name = f"page_{p.tab_number}_{p.subtab_number}"
        page_dict[page_name] = p
        
    data = await websocket.receive_json()
    PAGE = page_dict.get(data["page"])
    if PAGE:
        async for key, value in run_websocket_data(PAGE, data):
            v = value.dict()
            v1 = json.dumps(v, cls=myencoder)
            await websocket.send_json({key: v1})

async def run_func(d, wc,p):
    if d is None:
        v = await wc.view()
    else:
        v = await wc.view(d)
    return f"{wc.full_url()}$ZenLookup${p}", v

async def run_websocket_data(page:Type[ZenPage], data:dict):
    p = data["page"]
    if page.pydantic_class is None:
        calls = [ run_func(None, wc, p) for wc in page.websocket_calls]
    else:    
        d = page.pydantic_class(**data)
        calls = [ run_func(d, wc, p) for wc in page.websocket_calls]
    
    for coro in asyncio.as_completed(calls):
        result = await coro
        yield result



from typing import Optional, List, Union
from zen_dash.objects import ZenPage
import warnings
from zen_dash.support import BaseUpdate
from importlib.metadata import version
from pydantic import root_validator, validator

class SidebarTab(BaseUpdate):
    label: str
    icon: Optional[str] = 'Home'
    custom_url: Optional[str]

    

class SidebarGroup(BaseUpdate):
    name: str
    subtabs: List[SidebarTab]

class FilterInfo(BaseUpdate):
    url: str

class Sidebar(BaseUpdate):
   
    tabs: List[Union[SidebarTab, SidebarGroup]]
    filters: List[FilterInfo]
    size: str = '300px'
    library_version: str = version("zen-dash")

    @validator('tabs')
    def first_cant_be_group(cls, v):
        if len(v)> 1:
            if not isinstance(v[0], SidebarTab):
                raise Exception("First tab can't be SidebarGroup")
        return v


class GroupSidebarList(BaseUpdate):
    name: str
    sub_pages: List[ZenPage]

class SidebarList(BaseUpdate):
    pages: List[ZenPage]
    groupSidebar: List[GroupSidebarList]


def get_page_dict(pages: List[ZenPage]):
    page_dict = {}
    for p in pages:
        if p.tab_number is None:
            raise Exception("Tab number is not define in Zen Page")
        
        page_number = p.tab_number
        if p.subtab_number is None:
            
            page_dict[page_number] = p
        else:
            if page_dict.get(page_number) is None:
                page_dict[page_number] = {}
            elif isinstance(page_dict.get(page_number), dict):
                pass
            elif not isinstance(page_dict.get(page_number), dict):
                raise Exception(f"Please Check tab_number and subtab_number")
            
            sub_number = p.subtab_number
            page_dict[page_number][sub_number] = p
    return page_dict

def RenderTabs(sidebar_list: SidebarList):
    tabs = []
    
    for p in sidebar_list.pages:
        if p.subtab_number:
            raise Exception("Sidetab can't have subtab_number")
        
        if p.custom_url:
            url = p.custom_url
        else:
            url = f"page_{p.tab_number}"
        
        tabs.append(SidebarTab(label=p.name, icon=p.icon, custom_url=url))
    
    for subtab in sidebar_list.groupSidebar:
        sub_tab = []
        for sub_page in subtab.sub_pages:
            if sub_page.subtab_number is None:
                raise Exception(f"Please provde subpage for {sub_page.name}")
            if sub_page.custom_url:
                url = sub_page.custom_url
            else:
                url = f"page_{sub_page.tab_number}_{sub_page.subtab_number}"
            
            sub_tab.append(SidebarTab(label=sub_page.name, icon=sub_page.icon, custom_url=url))
        
        tabs.append(SidebarGroup(name= subtab.name, subtabs=sub_tab))

    return tabs







    


from pydantic import BaseModel

class BaseUpdate(BaseModel):
    def dict(
        self,
        *,
        include = None,
        exclude = None,
        by_alias: bool = False,
        skip_defaults: bool = None,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = True
    ) :
        return super().dict(
            include=include,
            exclude=exclude,
            by_alias=by_alias,
            skip_defaults=skip_defaults,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=True
        )


from typing import Dict, List, Optional, Type
from pydantic import BaseModel, root_validator

from zen_dash.objects.page import Page


class Zen:

    @staticmethod
    def full_url(*args, **kwargs) -> str:
        raise Exception("Not Implimented")

    @staticmethod
    def url(*args, **kwargs) -> str:
        raise Exception("Not Implimented")

    @staticmethod
    def view(*args, **kwargs):
        raise Exception("Not Implimented")

    @staticmethod
    def server_full_url(*args, **kwargs) -> str:
        raise Exception("Not Implimented")

    @staticmethod
    def server_url(*args, **kwargs) -> str:
        raise Exception("Not Implimented")

    @staticmethod
    def server_view(*args, **kwargs):
        raise Exception("Not Implimented")

    @staticmethod
    def server(*args, **kwargs):
        raise Exception("Not Implimented")


class ZenPage(BaseModel):
    name: str
    icon: str
    page: Page
    websocket_calls: List[Type[Zen]] = []
    pydantic_class: Type[BaseModel] = None
    tab_number: Optional[int]
    subtab_number: Optional[int]
    custom_url: Optional[str]

class RefreshInfo(BaseModel):
    refresh: bool = False
    rate_in_seconds: int =  5*60

class WebSocketConfig(BaseModel):
    active: bool = False

class Auth(BaseModel):
    SSO: bool = False
    UP_Login: bool = False
    

class Configuration(BaseModel):
    retry_count: int = 2
    show_right_sidebar: bool = False
    refresh: RefreshInfo =  RefreshInfo()
    websocket: WebSocketConfig = WebSocketConfig()
    auth: Auth = Auth() 




from typing import List
from zen_dash.objects import ZenPage
from zen_dash.objects.page import Instance, Row, Page


def get_page_dict(pages: List[ZenPage]):
    page_dict = {}
    custom_dict = {}
    for p in pages:
        if p.tab_number is None:
            raise Exception("Tab number is not define in Zen Page")
        
        if p.subtab_number is None:
            page_number = f"page_{p.tab_number}"
        else:
            page_number = f"page_{p.tab_number}_{p.subtab_number}"
        if page_dict.get(page_number):
            raise Exception(
                f"{page_dict.get(page_number)} and {p} pages are assing to same tab/subtab")

        else:
            page_dict[page_number] = p
        
        if p.custom_url is not None:
            custom_dict[p.custom_url] = p
    return page_dict, custom_dict

def RenderPage(pages: List[ZenPage], fragment: str):
    PAGEDICT, CUSTOMDICT = get_page_dict(pages)
    p = PAGEDICT.get(fragment)
    p1 = CUSTOMDICT.get(fragment)

    if p:
        return p.page
    elif p1:
        return p1.page
    else:
        raise Exception("Page is not define")
